import itertools
from collections import OrderedDict
from typing import List


def generate_permutations(tokens: List[dict]):
    """
    Initiate generate_permutations_recursive() on tokens
    with no processed_left_string.
    """
    return _generate_permutations_recursive(tokens, processed_left_string="", index=0)


def _generate_permutations_recursive(tokens, processed_left_string, index):
    """
    One by one sequentially, each input substructure generates permutations.
    Each permutation starts a new output instance.
    """
    if index >= len(tokens):  # no more input to process, exit recurency
        yield processed_left_string

    else:
        cursor_permutations = _permute(tokens[index])

        for permutation in cursor_permutations:
            yield from _generate_permutations_recursive(
                tokens, processed_left_string + permutation, index + 1
            )


def _permute(token):
    results = []

    permutations = itertools.permutations(token.items())
    for permutation in permutations:
        # serialize nested list-dict structure to a flat list of strings
        converted_permutation = [""]  # output structure: list of strings

        for key, value in permutation:
            if isinstance(value, str):
                pattern = [f'{key}: "{value}" ']
                converted_permutation = [
                    "".join(x)
                    for x in itertools.product(converted_permutation, pattern)
                ]
            elif isinstance(value, bool):
                pattern = [f"{key}: {value} "]
                converted_permutation = [
                    "".join(x)
                    for x in itertools.product(converted_permutation, pattern)
                ]
            elif isinstance(value, OrderedDict):
                recursive_permutation = _permute(value)
                prefix = [f" {key} {{ "]
                suffix = [" } "]
                converted_permutation = [
                    "".join(x)
                    for x in itertools.product(
                        converted_permutation, prefix, recursive_permutation, suffix
                    )
                ]
            else:
                raise ValueError(
                    "Permuter error! Substructure type is incorrect, something went really wrong."
                )

        results.extend(converted_permutation)

    return results



import logging

log_level = {"info": logging.INFO, "debug": logging.DEBUG}


class ITN_logger:
    def __init__(self, logger_name: str):
        self._logger = logging.getLogger(logger_name)
        self.console_handler = logging.StreamHandler()
        self.formatter = logging.Formatter(
            "[%(asctime)s][%(levelname)s][%(name)s] - %(message)s"
        )
        self.console_handler.setFormatter(self.formatter)
        self._logger.addHandler(self.console_handler)

    @property
    def logger(self):
        return self._logger

    def set_level(self, level: str):
        self.logger.setLevel(log_level[level])
        self.console_handler.setLevel(log_level[level])



import logging
import pynini
from pathlib import Path

from pl_itn.src.grammar import Grammar, GrammarType
from pl_itn.src.tag import tag
from pl_itn.src.parse import parse_tokens
from pl_itn.src.permute import generate_permutations
from pl_itn.src.restore_uppercase import restore_uppercase
from pl_itn.src.restore_whitespace import restore_whitespaces
from pl_itn.src.tag import tag
from pl_itn.src.utils import pre_process, post_process
from pl_itn.src.verbalize import verbalize

logger = logging.getLogger(__name__)

package_root = Path(__file__).parents[1]


class NormalizationError(Exception):
    ...


class Normalizer:
    def __init__(
        self,
        tagger_fst_path: Path = package_root / "grammars/tagger.fst",
        verbalizer_fst_path: Path = package_root / "grammars/verbalizer.fst",
    ):
        self._tagger = Grammar(tagger_fst_path, GrammarType.TAGGER)
        self._verbalizer = Grammar(verbalizer_fst_path, GrammarType.VERBALIZER)

    @property
    def tagger(self):
        return self._tagger

    @property
    def verbalizer(self):
        return self._verbalizer

    def set_grammar(
        self, grammar_fst_path: Path, grammar_type: GrammarType, description: str = ""
    ):
        if grammar_type == GrammarType.TAGGER:
            self._tagger = Grammar(grammar_fst_path, GrammarType.TAGGER, description)
        else:
            self._verbalizer = Grammar(
                grammar_fst_path, GrammarType.VERBALIZER, description
            )

    def __call__(self, text: str) -> str:
        return self.normalize(text)

    def normalize(self, text: str) -> str:
        logger.debug(f"input: {text}")

        preprocessed_text = pre_process(text)

        if not preprocessed_text:
            logger.info("Empty input string")
            return text

        logger.debug(f"pre_process(): {preprocessed_text}")

        try:
            tagged_text = tag(self.tagger.fst, preprocessed_text)
            logger.debug(f"tag(): {tagged_text}")

            tokens = parse_tokens(tagged_text)
            logger.debug(f"parse(): {tokens}")

            tags_reordered = generate_permutations(tokens)
            logger.debug(f"generate_permutations(): {tags_reordered}")

            verbalized_text = verbalize(self.verbalizer.fst, tags_reordered)
            logger.debug(f"verbalize(): {verbalized_text}")

            postprocessed_text = post_process(verbalized_text)
            logger.debug(f"post_process(): {postprocessed_text}")

            uppercase_restored = restore_uppercase(text, postprocessed_text)
            logger.debug(f"restore_uppercase(): {uppercase_restored}")

            whitespaces_restored = restore_whitespaces(text, uppercase_restored, tokens)
            logger.debug(f"restore_whitespaces(): {whitespaces_restored}")

            return whitespaces_restored

        except Exception as e:
            logger.error(e)
            raise NormalizationError(e)



import re
from collections import OrderedDict
from typing import List

"""
Parses tokenized/classified text, e.g. 
'tokens { money { integer: "20" currency: "$" } } tokens { name: "left"}'

Args
    text: tokenized text
"""


def parse_tokens(text) -> List[dict]:
    parsed_tokens = []

    chunks = text.split("tokens")
    chunks_with_length = [c for c in chunks if len(c)]

    for token in chunks_with_length:
        current_level_dict = OrderedDict()
        current_level_dict["tokens"] = _parse_recursively(token)
        parsed_tokens.append(current_level_dict)

    return parsed_tokens


def _strip_braces_and_whitespaces(text):
    text = text.strip()
    text = text.lstrip("{")
    text = text.rstrip("}")
    text = text.strip()
    return text


def _parse_final_leaf(text):
    text = _strip_braces_and_whitespaces(text)
    keywords = re.findall(r"(\S+): ", text)
    values = re.findall(r"\"([^\"]+)\"", text)

    if len(keywords) != len(values):
        raise ValueError("Parsing tagged text into tokens failed.")

    yield from zip(keywords, values)


def _parse_recursively(text):
    text = _strip_braces_and_whitespaces(text)
    current_level_dict = OrderedDict()

    if "{" not in text:  # no nested dict
        for keyword, value in _parse_final_leaf(text):
            current_level_dict[keyword] = value

    else:
        keyword, text = text.split("{", 1)
        keyword = keyword.strip()
        current_level_dict[keyword] = _parse_recursively(text)

    return current_level_dict



import logging
import pynini
import re

logger = logging.getLogger(__name__)


def verbalize(verbalize_fst, tags_permutations):
    for tagged_text_permutation in tags_permutations:
        
        tagged_text_permutation = re.sub(r" +", " ", tagged_text_permutation)
        tagged_text_permutation = tagged_text_permutation.strip()
        
        tagged_text_permutation = pynini.escape(tagged_text_permutation)
        logger.debug(tagged_text_permutation)
        lattice = tagged_text_permutation @ verbalize_fst

        if lattice.num_states():
            # permutation is syntactically correct (foreseen by verbalizer)
            return pynini.shortestpath(lattice, nshortest=1, unique=True).string()

    # no syntactically correct permutation found
    raise ValueError("Verbalizer returned no output")



import pynini


def tag(tagger_fst, text):
    text = pynini.escape(text)

    lattice = text @ tagger_fst

    text = pynini.shortestpath(lattice, nshortest=1, unique=True).string()

    if text is None:
        raise ValueError("Tagger returned no output.")
    return text



import argparse
import yaml
from pathlib import Path
from signal import signal, SIGINT, SIGTERM

from pl_itn import Normalizer, package_root
from pl_itn.exceptions import InterruptException, gentle_interrupt_handler
from pl_itn.logging import ITN_logger

def main():
    args = parser()

    itn_logger = ITN_logger("pl_itn")
    itn_logger.set_level(args.log_level)

    if args.config:
        with args.config.open() as config_file:
            config = yaml.safe_load(config_file)
        args.tagger = Path(config.get("out_dir")) / config.get("tagger_fname")
        args.verbalizer = Path(config.get("out_dir")) / config.get("verbalizer_fname")

    normalizer = Normalizer(
        tagger_fst_path=args.tagger,
        verbalizer_fst_path=args.verbalizer,
    )

    if args.interactive:
        run_interactive(normalizer)
    else:
        run_single(normalizer, args.text)


def parser():
    parser = argparse.ArgumentParser(
        description="Inverse Text Normalization based on Finite State Transducers"
    )

    input_args = parser.add_mutually_exclusive_group(required=True)
    input_args.add_argument("-t", "--text", type=str, help="Input text")
    input_args.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        help="If used, demo will process phrases from stdin interactively.",
    )

    parser.add_argument(
        "--tagger", type=Path, default=(package_root / "grammars/tagger.fst")
    )
    parser.add_argument(
        "--verbalizer", type=Path, default=(package_root / "grammars/verbalizer.fst")
    )
    parser.add_argument(
        "--config",
        type=Path,
        help="Optionally provide yaml config with tagger and verbalizer paths.",
    )
    parser.add_argument("--log-level", choices=["debug", "info"], default="info")

    return parser.parse_args()


def run_interactive(normalizer: Normalizer):
    signal(SIGINT, gentle_interrupt_handler)  # keyboard interrupt
    signal(SIGTERM, gentle_interrupt_handler)  # terminal interrupt

    print("Enter your phrase:")

    while True:
        try:
            phrase = input()
            print(normalizer.normalize(phrase))
            print()
        except InterruptException:
            break


def run_single(normalizer: Normalizer, text: str) -> None:
    print(normalizer.normalize(text))


if __name__ == "__main__":
    main()



import re
import string
from typing import List, Dict


def restore_whitespaces(original: str, normalized: str, tokens: List[Dict]) -> str:
    """
    Finite State Transducers normalization loses record of the original whitespace.
    Each token of the normalized text is separated by a single space, no matter if there was no, single or multiple spaces in the given spot originally.
            `! (dwunasta)` -> `! ( 12 )`

    WhitespaceRestorer detects words which have not been transformed by the normalizer (mostly punctuation marks) and restores the original spaces around them.
            `! (dwunasta)` -> `! ( 12 )` -> `! (12)`

    Args:
        original (str): Input phrase, for example "Jest godzina piętnasta trzydzieści."
        normalized (str): ITN output phrase, for example "Jest godzina 15:30 ."
        tokens ([dict]): parsed tokens from FST tagging, for example
            `[OrderedDict([('tokens', OrderedDict([('name', 'jest')]))]), OrderedDict([('tokens', OrderedDict([('time', OrderedDict([('complement', 'godzina'), ('hours', '15'), ('minutes', '30')]))]))]), OrderedDict([('tokens', OrderedDict([('name', '.')]))])]`
    """
    lower_list = lambda x: [w.lower() for w in x]

    normalized = normalized.split(" ")
    normalized_lower = lower_list(normalized)

    original = re.split(
        rf"(\s+|[{string.punctuation}])", original
    )  # ie. ['Jest', ' ', 'godzina', ' ', 'piętnasta', ' ', 'trzydzieści', '.']
    original = [w for w in original if len(w) > 0]
    original_lower = lower_list(original)

    restored = ""

    for token in tokens:
        token = token[
            "tokens"
        ]  # OrderedDict([('tokens', OrderedDict([('name', 'jest')]))]) -> OrderedDict([('name', 'jest')])

        # Tokens tagged as `name` have not been transformed during FST normalization.
        # ie. `OrderedDict([('name', 'jest')])` or `OrderedDict([('name', '.')])`
        is_original_form = lambda x: x.get("name")

        if is_original_form(token):
            word = token["name"]

            # the word has not been transformed, so it can be found both in the normalized and original texts
            index_normalized = normalized_lower.index(word)
            index_original = original_lower.index(word)

            # if the word is preceded by transformed words in the normalized text, add them first
            restored += " ".join(normalized[:index_normalized])

            # compose a group of the word itself sorrounded by its original white spaces. Add it to the output.
            def maybe_get_whitespace(input_list, index):
                try:
                    value = input_list[index]
                except IndexError:
                    return ""
                if re.match("^\s+$", value):
                    return value
                return ""

            word_with_spaces = (
                maybe_get_whitespace(original, index_original - 1)
                + original[index_original]
                + maybe_get_whitespace(original, index_original + 1)
            )
            restored += word_with_spaces

            if maybe_get_whitespace(original, index_original + 1):
                # Include trailing whitespace to the processed part
                index_original += 1

            # Remove processed input
            def slice_head(input_list, index):
                return input_list[index + 1 :]

            normalized = slice_head(normalized, index_normalized)
            normalized_lower = slice_head(normalized_lower, index_normalized)
            original = slice_head(original, index_original)
            original_lower = slice_head(original_lower, index_original)

    restored += " ".join(normalized)
    return restored



import re
from string import punctuation


def pre_process(text: str):
    text = text.lower()
    text = _surround_punctuation_with_spaces(text)
    text = _remove_multiple_whitespaces(text)
    return text


def post_process(text: str):
    text = _remove_multiple_whitespaces(text)
    return text


def _surround_punctuation_with_spaces(text: str):
    for punct in punctuation:
        text = text.replace(punct, f" {punct} ")
    return text


def _remove_multiple_whitespaces(text: str):
    text = re.sub(r"\s+", " ", text)
    text = text.strip()
    return text



def restore_uppercase(original: str, normalized: str):
    original = original.split()
    normalized = normalized.split()
    restored = normalized[:]

    original_index, original_limit = 0, len(original)
    normalized_index, normalized_limit = 0, len(normalized)

    def finished():
        return normalized_index >= normalized_limit or original_index >= original_limit

    def match():
        if finished():  # out of index guardian
            return False
        return normalized[normalized_index] == original[original_index].lower()

    def next_original_index():
        nonlocal original_index
        original_index += 1
        return not finished()  # False if index out of range

    def next_normalized_index():
        nonlocal normalized_index
        normalized_index += 1

    def restore():
        restored[normalized_index] = original[original_index]
        next_original_index()
        next_normalized_index()

    while not finished():
        if match():
            restore()
            break

        # check the next pair
        next_original_index()
        next_normalized_index()
        if match():
            restore()
            break

        # look for shifted match
        checkpoint = original_index
        while next_original_index():
            if match():
                restore()
                break

        # no match, the word has been inverse normalized
        original_index = checkpoint
        next_normalized_index()
        next_original_index()

    return " ".join(restored)



from enum import Enum
from pathlib import Path

from pynini import Fst, FstIOError

GrammarType = Enum("GrammarType", ["TAGGER", "VERBALIZER"])


class Grammar:
    def __init__(
        self, fst_path: Path, grammar_type: GrammarType, description: str = ""
    ):
        # try:
        self._fst = Fst.read(str(fst_path))
        self._type = grammar_type
        # except FstIOError as e:
        self._fst_path = fst_path
        self._description = description

    @property
    def fst(self):
        return self._fst

    @property
    def grammar_type(self):
        return self._type

    @property
    def fst_path(self):
        return self._fst_path

    @property
    def description(self):
        return self._description
